import rio
import rio.components.layout_display
import asyncio
import sys
from typing import Any
import uniserde


class RootComponent(rio.Component):
    def on_change(self, new_component_id) -> None:
        print("Changed", new_component_id)

    def build(self) -> rio.Component:
        target = rio.Row(
            rio.Text("Hey There"),
            rio.Button("Click Me"),
            spacing=3,
            margin_left=5,
            margin_top=2,
            height=10,
            # align_x=0.5,
            # align_y=0.5,
        )

        return rio.Column(
            target,
            # rio.components.layout_display.LayoutDisplay(
            #     component_id=target._id,
            #     on_component_change=self.on_change,
            #     width=60,
            #     height=40,
            #     align_x=0.5,
            #     align_y=0.5,
            # ),
        )


app = rio.App(
    build=RootComponent,
    default_attachments=[],
)



import asyncio

import rio.testing


class ChildToggler(rio.Component):
    child: rio.Component
    switch: bool = True

    def toggle(self) -> None:
        self.switch = not self.switch

    def build(self) -> rio.Component:
        if self.switch:
            return rio.Spacer()
        else:
            return self.child


async def test_mounted():
    mounted = unmounted = False

    class DemoComponent(rio.Component):
        @rio.event.on_mount
        def _on_mount(self):
            nonlocal mounted
            mounted = True

        @rio.event.on_unmount
        def _on_unmount(self):
            nonlocal unmounted
            unmounted = True

        def build(self) -> rio.Component:
            return NestedComponent()

    class NestedComponent(rio.Component):
        def build(self) -> rio.Component:
            return rio.Text("hi")

    def build():
        return ChildToggler(DemoComponent())

    async with rio.testing.TestClient(build) as test_client:
        root = test_client.get_component(ChildToggler)
        assert not mounted
        assert not unmounted

        root.toggle()
        await test_client.refresh()
        assert mounted
        assert not unmounted

        # Make sure the newly mounted components were sent to the client
        demo_component = test_client.get_component(DemoComponent)
        nested_component = test_client.get_component(NestedComponent)
        text_component = test_client.get_component(rio.Text)
        assert test_client._last_updated_components == {
            root,
            demo_component,
            nested_component,
            text_component,
        }

        root.toggle()
        await test_client.refresh()
        assert unmounted


async def test_refresh_after_synchronous_mount_handler():
    class DemoComponent(rio.Component):
        mounted: bool = False

        @rio.event.on_mount
        def on_mount(self):
            self.mounted = True

        def build(self) -> rio.Component:
            return rio.Switch(self.mounted)

    async with rio.testing.TestClient(DemoComponent) as test_client:
        demo_component = test_client.get_component(DemoComponent)
        switch = test_client.get_component(rio.Switch)

        # TODO: I don't know how we can wait for the refresh, so I'll just use a
        # sleep()
        await asyncio.sleep(0.5)
        assert demo_component.mounted

        last_component_state_changes = test_client._last_component_state_changes
        assert switch in last_component_state_changes
        assert last_component_state_changes[switch].get("is_on") is True



from __future__ import annotations

import abc
import asyncio
import json
import logging
import time
import warnings
import weakref
from datetime import date
from pathlib import Path
from typing import *

import langcodes
import pytz
import starlette.datastructures

import rio
import rio.assets

from .. import (
    assets,
    data_models,
    language_info,
    routing,
    session,
    user_settings_module,
    utils,
)
from ..transports import (
    AbstractTransport,
    TransportClosedIntentionally,
    TransportInterrupted,
)

__all__ = ["AbstractAppServer"]


class AbstractAppServer(abc.ABC):
    def __init__(
        self,
        app: rio.App,
        *,
        running_in_window: bool,
        debug_mode: bool,
    ):
        self.app = app
        self.running_in_window = running_in_window
        self.debug_mode = debug_mode

        self._session_serve_tasks = dict[rio.Session, asyncio.Task[object]]()
        self._disconnected_sessions = dict[rio.Session, float]()

        # If running as a server, periodically clean up expired sessions
        if not running_in_window:
            asyncio.create_task(
                _periodically_clean_up_expired_sessions(weakref.ref(self)),
                name="Periodic session cleanup",
            )

    @abc.abstractmethod
    async def file_chooser(
        self,
        session: rio.Session,
        *,
        file_extensions: Iterable[str] | None = None,
        multiple: bool = False,
    ) -> utils.FileInfo | tuple[utils.FileInfo, ...]:
        raise NotImplementedError

    @abc.abstractmethod
    def url_for_user_asset(self, relative_asset_path: Path) -> rio.URL:
        """
        Returns the URL of an asset from the app's `assets_dir`. This must be a
        permalink, i.e. a URL that doesn't change if the server is restarted.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def weakly_host_asset(self, asset: assets.HostedAsset) -> str:
        """
        Hosts an asset as long as it is alive. Returns the asset's URL as a
        string.
        """
        raise NotImplementedError

    def host_asset_with_timeout(
        self, asset: assets.HostedAsset, timeout: float
    ) -> str:
        """
        Hosts an asset for a limited time. Returns the asset's URL as a string.
        """
        url = self.weakly_host_asset(asset)

        async def keep_alive():
            await asyncio.sleep(timeout)
            _ = asset

        asyncio.create_task(
            keep_alive(), name=f"Keep asset {asset} alive for {timeout} sec"
        )

        return url

    def _after_session_closed(self, session: rio.Session) -> None:
        """
        Called by `Session.close()`. Gives the server an opportunity to clean
        up.
        """
        # Stop the task that's listening for incoming messages
        task = self._session_serve_tasks.pop(session)
        task.cancel()

    async def create_session(
        self,
        initial_message: data_models.InitialClientMessage | None,
        *,
        transport: AbstractTransport,
        url: rio.URL,
        client_ip: str,
        client_port: int,
        http_headers: starlette.datastructures.Headers,
    ) -> rio.Session:
        """
        Creates a new session.

        ## Raises

        `NavigationFailed`: If a page guard crashes
        """
        if initial_message is None:
            initial_message = data_models.InitialClientMessage.from_defaults()

        # Normalize and deduplicate the languages
        preferred_languages: list[str] = []

        for language in initial_message.preferred_languages:
            try:
                language = langcodes.standardize_tag(language)
            except ValueError:
                continue

            if language not in preferred_languages:
                preferred_languages.append(language)

        if len(preferred_languages) == 0:
            preferred_languages.append("en-US")

        # Get locale information
        if len(initial_message.decimal_separator) != 1:
            logging.warning(
                f'Client sent invalid decimal separator "{initial_message.decimal_separator}". Using "." instead.'
            )
            initial_message.decimal_separator = "."

        if len(initial_message.thousands_separator) > 1:
            logging.warning(
                f'Client sent invalid thousands separator "{initial_message.thousands_separator}". Using "" instead.'
            )
            initial_message.thousands_separator = ""

        # There does not seem to be any good way to determine the first day of
        # the week in JavaScript. Look up the first day of the week based on
        # the preferred language.
        first_day_of_week = language_info.get_week_start_day(
            preferred_languages[0]
        )

        # Make sure the date format string is valid
        try:
            formatted_date = date(3333, 11, 22).strftime(
                initial_message.date_format_string
            )
        except ValueError:
            date_format_string_is_valid = False
        else:
            date_format_string_is_valid = (
                "33" in formatted_date
                and "11" in formatted_date
                and "22" in formatted_date
            )

        if not date_format_string_is_valid:
            logging.warning(
                f'Client sent invalid date format string "{initial_message.date_format_string}". Using "%Y-%m-%d" instead.'
            )
            initial_message.date_format_string = "%Y-%m-%d"

        # Parse the timezone
        try:
            timezone = pytz.timezone(initial_message.timezone)
        except pytz.UnknownTimeZoneError:
            logging.warning(
                f'Client sent unknown timezone "{initial_message.timezone}". Using UTC instead.'
            )
            timezone = pytz.UTC

        base_url = url.with_path("").with_query("").with_fragment("")

        # Set the theme according to the user's preferences
        theme = self.app._theme
        if isinstance(theme, tuple):
            if initial_message.prefers_light_theme:
                theme = theme[0]
            else:
                theme = theme[1]

        # Prepare the initial URL. This will be exposed to the session as the
        # `active_page_url`, but overridden later once the page guards have been
        # run.
        initial_page_url = url

        # Create the session
        sess = session.Session(
            app_server_=self,
            transport=transport,
            client_ip=client_ip,
            client_port=client_port,
            http_headers=http_headers,
            base_url=base_url,
            active_page_url=initial_page_url,
            timezone=timezone,
            preferred_languages=preferred_languages,
            month_names_long=initial_message.month_names_long,
            day_names_long=initial_message.day_names_long,
            date_format_string=initial_message.date_format_string,
            first_day_of_week=first_day_of_week,
            decimal_separator=initial_message.decimal_separator,
            thousands_separator=initial_message.thousands_separator,
            window_width=initial_message.window_width,
            window_height=initial_message.window_height,
            theme_=theme,
        )

        # Deserialize the user settings
        await sess._load_user_settings(initial_message.user_settings)

        # Add any remaining attachments
        for attachment in self.app.default_attachments:
            if not isinstance(attachment, user_settings_module.UserSettings):
                sess.attach(attachment)

        # Start listening for incoming messages. This should happen before
        # `on_session_start` is called, so that we don't deadlock in case
        # someone calls a method that requires a response from the client.
        self._session_serve_tasks[sess] = asyncio.create_task(
            self._serve_session(sess)
        )

        # Trigger the `on_session_start` event.
        #
        # Since this event is often used for important initialization tasks like
        # adding attachments, actually wait for it to finish before continuing.
        #
        # However, also don't run it too early, since this function expects a
        # (mostly) functioning session.
        #
        # TODO: Figure out which values are still missing, and expose them,
        #       expose placeholders, or document that they aren't available.
        start_time = time.monotonic()
        await sess._call_event_handler(
            self.app._on_session_start,
            sess,
            refresh=False,
        )
        duration = time.monotonic() - start_time

        if duration > 5:
            warnings.warn(
                f"Session startup was delayed for {duration:.0f} seconds by"
                f" `on_session_start`. If you have long-running operations that"
                f" don't have to finish before the session starts, you should"
                f" execute them in a background task."
            )

        # Run any page guards for the initial page. Throws a `NavigationFailed`
        # if a page guard crashed.
        (
            active_page_instances,
            active_page_url_absolute,
        ) = routing.check_page_guards(sess, initial_page_url)

        # Is this a page, or a full URL to another site?
        try:
            utils.make_url_relative(
                sess._base_url,
                active_page_url_absolute,
            )

        # This is an external URL. Navigate to it
        except ValueError:

            async def history_worker() -> None:
                await sess._evaluate_javascript(
                    f"""
                    window.location.href = {json.dumps(str(active_page_url_absolute))};
                    """
                )

            sess.create_task(history_worker(), name="navigate to external URL")

            # TODO: End the session? Abort initialization?

        # Set the initial page URL. When connecting to the server, all relevant
        # page guards execute. These may change the URL of the page, so the
        # client needs to take care to update the browser's URL to match the
        # server's.
        if active_page_url_absolute != initial_page_url:

            async def update_url_worker():
                js_page_url = json.dumps(str(active_page_url_absolute))
                await sess._evaluate_javascript(
                    f"""
                    console.trace("Updating browser URL to match the one modified by guards:", {js_page_url});
                    window.history.replaceState(null, "", {js_page_url});
                    """
                )

            sess.create_task(
                update_url_worker(),
                name="Update browser URL to match the one modified by guards",
            )

        # Update the session's active page and instances
        sess._active_page_instances = tuple(active_page_instances)
        sess._active_page_url = active_page_url_absolute

        # Apply the CSS for the chosen theme
        await sess._apply_theme(theme)

        # Send the first `updateComponentStates` message
        await sess._refresh()

        return sess

    async def _serve_session(self, sess: rio.Session) -> None:
        try:
            await sess.serve()
        except TransportClosedIntentionally:
            sess.close()
        except TransportInterrupted:
            # Connection was interrupted, mark the session as disconnected but
            # keep it alive for a while to see if the client reconnects
            sess._transport = None


async def _periodically_clean_up_expired_sessions(
    app_server_ref: weakref.ReferenceType[AbstractAppServer],
):
    LOOP_INTERVAL = 60 * 15
    SESSION_LIFETIME = 60 * 60

    while True:
        await asyncio.sleep(LOOP_INTERVAL)

        app_server = app_server_ref()
        if app_server is None:
            return

        now = time.monotonic()
        cutoff = now - SESSION_LIFETIME

        disconnected_sessions = list(app_server._disconnected_sessions.items())
        for sess, disconnect_time in disconnected_sessions:
            if disconnect_time < cutoff:
                sess.close()

        # Drop the reference to the app server
        app_server = None



from __future__ import annotations

from dataclasses import KW_ONLY
from typing import *  # type: ignore

from uniserde import JsonDoc

import rio

from .fundamental_component import FundamentalComponent

__all__ = [
    "LayoutDisplay",
]


@final
class LayoutDisplay(FundamentalComponent):
    component_id: int

    _: KW_ONLY

    on_component_change: rio.EventHandler[int] = None

    def _validate_delta_state_from_frontend(self, delta_state: JsonDoc) -> None:
        if not set(delta_state) <= {"component_id"}:
            raise AssertionError(
                f"Frontend tried to change `{type(self).__name__}` state: {delta_state}"
            )

    async def _call_event_handlers_for_delta_state(
        self, delta_state: JsonDoc
    ) -> None:
        # Trigger on_change event
        try:
            new_value = delta_state["component_id"]
        except KeyError:
            pass
        else:
            assert isinstance(new_value, int), new_value
            await self.call_event_handler(self.on_component_change, new_value)


LayoutDisplay._unique_id = "LayoutDisplay-builtin"



from __future__ import annotations

import io
from dataclasses import KW_ONLY
from typing import *  # type: ignore

import rio


class AlignmentControls(rio.Component):
    label: str
    enabled: bool
    value: float

    def build(self) -> rio.Component:
        # Allow enabling / disabling alignment
        result = rio.Row(
            rio.Text(
                self.label,
                align_x=0,
                width="grow",
            ),
            rio.Switch(is_on=self.bind().enabled),
            spacing=1,
        )

        # If enabled, add a slider to control the alignment value
        if self.enabled:
            result = rio.Column(
                result,
                rio.Slider(
                    value=self.bind().value,
                    minimum=0,
                    maximum=1,
                ),
                spacing=0.5,
            )

        return result


class LayoutPreview(rio.Component):
    component: rio.Component

    _: KW_ONLY

    _width: float = 0
    _height: float = 0

    _margin_left: float = 0
    _margin_top: float = 0
    _margin_right: float = 0
    _margin_bottom: float = 0

    _enable_align_x: bool = False
    _enable_align_y: bool = False

    _align_x: float = 0.5
    _align_y: float = 0.5

    def _build_result(self) -> rio.Component:
        child_rectangle = rio.Card(
            content=rio.Text(type(self.component).__name__),
            color="hud",
            width=self._width,
            height=self._height,
            margin_left=self._margin_left,
            margin_top=self._margin_top,
            margin_right=self._margin_right,
            margin_bottom=self._margin_bottom,
        )

        margin_rectangle = rio.Rectangle(
            content=child_rectangle,
            fill=self.session.theme.neutral_palette.foreground.replace(
                opacity=0.2
            ),
            corner_radius=1,
            align_x=self._align_x if self._enable_align_x else None,
            align_y=self._align_y if self._enable_align_y else None,
        )

        return rio.Rectangle(
            content=margin_rectangle,
            fill=self.session.theme.neutral_palette.background_variant,
            height=20,
            corner_radius=1,
            width="grow",
        )

    def _build_source_code(self) -> rio.Component:
        source = io.StringIO()
        source.write(
            """SampleComponent(
    ...
"""
        )

        # Size
        if self._width != 0:
            source.write(f"    width={self._width:.1f},\n")

        if self._height != 0:
            source.write(f"    height={self._height:.1f},\n")

        # Margins
        single_x_margin = self._margin_left == self._margin_right
        single_y_margin = self._margin_top == self._margin_bottom

        margins: list[tuple[str, float]] = []

        if single_x_margin and single_y_margin:
            margins.append(("margin", self._margin_left))

        else:
            if single_x_margin:
                margins.append(("margin_x", self._margin_left))
            else:
                margins.append(("margin_left", self._margin_left))
                margins.append(("margin_right", self._margin_right))

            if single_y_margin:
                margins.append(("margin_y", self._margin_top))
            else:
                margins.append(("margin_top", self._margin_top))
                margins.append(("margin_bottom", self._margin_bottom))

        for margin_name, margin_value in margins:
            margin_value = round(margin_value, 1)

            if margin_value != 0:
                source.write(f"    {margin_name}={margin_value},\n")

        # Alignment
        if self._enable_align_x:
            source.write(f"    align_x={self._align_x:.2f},\n")

        if self._enable_align_y:
            source.write(f"    align_y={self._align_y:.2f},\n")

        source.write(")")

        # Display the source code
        return rio.CodeBlock(
            source.getvalue(),
            language="python",
        )

    def _build_explanations(self) -> rio.Component:
        # Some messages depend on data about the component
        COMPONENT_WIDTH = 7.6
        COMPONENT_HEIGHT = 1

        # Find all messages to display
        explanations: List[str] = []

        if not self._enable_align_x and not self._enable_align_y:
            explanations.append(
                "The component is using the full size of its parent because it doesn't have any alignment set."
            )

        else:
            if self._enable_align_x:
                if self._width < COMPONENT_WIDTH:
                    explanations.append(
                        "The configured width is less than the component's natural width. Because of this, the natural width of the component is used."
                    )
                else:
                    explanations.append(
                        "The component's width was overridden by your settings."
                    )
            else:
                explanations.append(
                    "The component is using the full width of its parent because it doesn't have any horizontal alignment set."
                )

            if self._enable_align_y:
                if self._height < COMPONENT_HEIGHT:
                    explanations.append(
                        "The configured height is less than the component's natural height. Because of this, the natural height of the component is used."
                    )
                else:
                    explanations.append(
                        "The component's height was overridden by your settings."
                    )
            else:
                explanations.append(
                    "The component is using the full height of its parent because it doesn't have any vertical alignment set."
                )

        # Combine everything
        if explanations:
            return rio.Column(
                # rio.Text("Explanations", style="heading3"),
                *[
                    rio.Text(
                        explanation,
                        justify="left",
                        wrap=True,
                        style=rio.TextStyle(
                            italic=True,
                        ),
                    )
                    for explanation in explanations
                ],
                spacing=0.5,
                margin_y=0.5,
            )

        return rio.Spacer(height=0)

    def _build_controls(self) -> rio.Component:
        controls_margin = rio.Column(
            rio.Text("Margin", style="heading3"),
            rio.Grid(
                [
                    rio.Text("Left"),
                    rio.Slider(
                        value=self.bind()._margin_left,
                        minimum=0,
                        maximum=5,
                        width="grow",
                    ),
                    rio.Text(f"{self._margin_left:.1f}"),
                ],
                [
                    rio.Text("Top"),
                    rio.Slider(
                        value=self.bind()._margin_top,
                        minimum=0,
                        maximum=5,
                        width="grow",
                    ),
                    rio.Text(f"{self._margin_top:.1f}"),
                ],
                [
                    rio.Text("Right"),
                    rio.Slider(
                        value=self.bind()._margin_right,
                        minimum=0,
                        maximum=5,
                        width="grow",
                    ),
                    rio.Text(f"{self._margin_right:.1f}"),
                ],
                [
                    rio.Text("Bottom"),
                    rio.Slider(
                        value=self.bind()._margin_bottom,
                        minimum=0,
                        maximum=5,
                        width="grow",
                    ),
                    rio.Text(f"{self._margin_bottom:.1f}"),
                ],
                row_spacing=1,
                column_spacing=1,
                margin=0.8,
            ),
            spacing=1,
            align_y=0,
        )

        controls_size = rio.Column(
            rio.Text("Size", style="heading3"),
            rio.Grid(
                [
                    rio.Text("Width"),
                    rio.Slider(
                        value=self.bind()._width,
                        minimum=0,
                        maximum=30,
                        width="grow",
                    ),
                    rio.Text(
                        f"{self._width:.1f}",
                        width=3.0,
                    ),
                ],
                [
                    rio.Text("Height"),
                    rio.Slider(
                        value=self.bind()._height,
                        minimum=0,
                        maximum=10,
                        width="grow",
                    ),
                    rio.Text(
                        f"{self._height:.1f}",
                        width=3.0,
                    ),
                ],
                row_spacing=0.5,
                column_spacing=0.5,
                margin=0.8,
            ),
            spacing=0.5,
            align_y=0,
        )

        controls_alignment = rio.Column(
            rio.Text("Alignment", style="heading3"),
            rio.Column(
                AlignmentControls(
                    label="Align X",
                    enabled=self.bind()._enable_align_x,
                    value=self.bind()._align_x,
                ),
                AlignmentControls(
                    label="Align Y",
                    enabled=self.bind()._enable_align_y,
                    value=self.bind()._align_y,
                ),
                spacing=0.5,
                margin=0.8,
            ),
            spacing=0.5,
            align_y=0,
        )

        return rio.Row(
            controls_margin,
            controls_size,
            controls_alignment,
            proportions="homogeneous",
            spacing=0.5,
        )

    def build(self) -> rio.Component:
        return rio.Column(
            self._build_result(),
            self._build_explanations(),
            self._build_controls(),
            self._build_source_code(),
            spacing=0.5,
            margin=1,
        )



import asyncio
import json
from collections.abc import Callable

from uniserde import JsonDoc

from .abstract_transport import *

__all__ = ["MessageRecorderTransport"]


class MessageRecorderTransport(AbstractTransport):
    def __init__(
        self, *, process_sent_message: Callable[[JsonDoc], None] | None = None
    ):
        super().__init__()

        self.process_sent_message = process_sent_message

        self.sent_messages = list[JsonDoc]()
        self._responses = asyncio.Queue[JsonDoc | None]()

    async def send(self, msg: str) -> None:
        parsed_msg = json.loads(msg)
        self.sent_messages.append(parsed_msg)

        if self.process_sent_message is not None:
            self.process_sent_message(parsed_msg)

    async def receive(self) -> JsonDoc:
        response = await self._responses.get()

        if response is None:
            raise TransportClosedIntentionally

        return response

    def close(self) -> None:
        self._responses.put_nowait(None)
        self.closed.set()

    def queue_response(self, response: JsonDoc) -> None:
        self._responses.put_nowait(response)



from pathlib import Path
from typing import *  # type: ignore

import rio
import rio.docs

from ... import utils
from . import layout_preview

try:
    RIO_PATH = Path(rio.__file__).parent
except Exception:
    # Use a dummy path. It doesn't really matter, as long as it doesn't crash
    # `Path.is_relative_to`
    RIO_PATH = Path.cwd() / "rio"


class ComponentDetails(rio.Component):
    component_id: int

    component_natural_width: float = 0
    component_natural_height: float = 0

    component_allocated_width: float = 0
    component_allocated_height: float = 0

    @rio.event.on_populate
    async def fetch_client_side_details(self) -> None:
        """
        Fetch additional details about the component which are only known
        client-side. Stores them in the class.
        """
        # Remember which component the details are being fetched for
        component_id = self.component_id

        # Fetch the details
        (allocation,) = await self.session._get_component_layouts(
            [component_id]
        )

        # If the current component has changed while the values were fetched,
        # don't update the state
        if self.component_id != component_id:
            return

        # Publish the results
        self.component_natural_width = allocation.natural_width
        self.component_natural_height = allocation.natural_height
        self.component_allocated_width = allocation.allocated_width
        self.component_allocated_height = allocation.allocated_height

    def _get_effective_margins(
        self, target: rio.Component
    ) -> Tuple[float, float, float, float]:
        return (
            utils.first_non_null(
                target.margin_left,
                target.margin_x,
                target.margin,
                0,
            ),
            utils.first_non_null(
                target.margin_top,
                target.margin_y,
                target.margin,
                0,
            ),
            utils.first_non_null(
                target.margin_right,
                target.margin_x,
                target.margin,
                0,
            ),
            utils.first_non_null(
                target.margin_bottom,
                target.margin_y,
                target.margin,
                0,
            ),
        )

    def build(self) -> rio.Component:
        # Get the target component
        try:
            target = self.session._weak_components_by_id[self.component_id]

        # In rare cases components can't be found, because they're created
        # client-side (e.g. injected margins). In this case, just don't display
        # anything.
        except KeyError:
            return rio.Spacer(height=0)

        debug_details = target._get_debug_details()

        return rio.Column(
            self._create_header(target),
            *self._create_instantiation_info(target),
            self._create_layout_details(target, debug_details),
            *self._create_extra_details(debug_details),
            # self._create_layout_revealer(target),
            spacing=0.2,
        )

    def _create_header(self, target: rio.Component) -> rio.Component:
        # Any values which should be displayed right in the title
        header_accessories = []

        if target.key is not None:
            header_accessories = [
                rio.Icon("material/key", fill="dim"),
                rio.Text(
                    target.key,
                    style="dim",
                    justify="left",
                ),
            ]

        # Link to docs
        if type(target)._rio_builtin_:
            docs_url = rio.docs.build_documentation_url(type(target).__name__)
            link_color = self.session.theme.secondary_color

            docs_link = [
                rio.Link(
                    rio.Row(
                        rio.Icon("material/library-books", fill=link_color),
                        rio.Text("Docs", style=rio.TextStyle(fill=link_color)),
                        spacing=0.5,
                        align_x=0,
                    ),
                    docs_url,
                    # TODO: Support icons in links
                    open_in_new_tab=True,
                    margin_left=0.5,
                ),
            ]
        else:
            docs_link = []

        return rio.Row(
            rio.Text(
                type(target).__qualname__,
                style="heading3",
            ),
            rio.Spacer(),
            *header_accessories,
            *docs_link,
            spacing=0.5,
        )

    def _create_instantiation_info(
        self, target: rio.Component
    ) -> Iterable[rio.Component]:
        # Which file/line was this component instantiated from?
        file, line = target._creator_stackframe_

        # If it was instantiated somewhere in the rio internals, don't show it
        if file.is_relative_to(RIO_PATH):
            return

        try:
            file = file.relative_to(Path.cwd())
        except ValueError:
            pass

        yield rio.Text(
            f"{file} line {line}",
            style="dim",
            justify="left",
        )

    def _create_layout_details(
        self, target: rio.Component, debug_details: dict[str, Any]
    ) -> rio.Component:
        # Create a grid with all the details
        grid = DetailsGrid()

        # Add some extra spacing
        grid.row += 1

        # Margins
        (
            margin_left,
            margin_top,
            margin_right,
            margin_bottom,
        ) = self._get_effective_margins(target)

        single_x_margin = margin_left == margin_right
        single_y_margin = margin_top == margin_bottom

        if single_x_margin and single_y_margin:
            grid.add_label("margin", column=0)
            grid.add_value(str(margin_left), column=1)

            grid.row += 1

        else:
            if single_x_margin:
                grid.add_label("margin_x", column=0)
                grid.add_value(str(margin_left), column=1)

            else:
                grid.add_label("margin_left", column=0)
                grid.add_value(str(margin_left), column=1)

                grid.add_label("margin_right", column=2)
                grid.add_value(str(margin_right), column=3)

            grid.row += 1

            if single_y_margin:
                grid.add_label("margin_y", column=0)
                grid.add_value(str(margin_top), column=1)

            else:
                grid.add_label("margin_top", column=0)
                grid.add_value(str(margin_top), column=1)

                grid.add_label("margin_bottom", column=2)
                grid.add_value(str(margin_bottom), column=3)

            grid.row += 1

        # Alignment
        if "align_x" in debug_details or "align_y" in debug_details:
            grid.add_label("align_x", column=0)
            grid.add_value(str(debug_details.get("align_x", "-")), column=1)

            grid.add_label("align_y", column=2)
            grid.add_value(str(debug_details.get("align_y", "-")), column=3)

            grid.row += 1

        # Size
        if "width" in debug_details or "height" in debug_details:
            try:
                py_width_str = debug_details["width"]
            except KeyError:
                py_width_str = "-"
            else:
                if isinstance(py_width_str, (int, float)):
                    py_width_str = round(py_width_str, 2)

                py_width_str = repr(py_width_str)

            try:
                py_height_str = debug_details["height"]
            except KeyError:
                py_height_str = "-"
            else:
                if isinstance(py_height_str, (int, float)):
                    py_height_str = round(py_height_str, 2)

                py_height_str = repr(py_height_str)

            # Add some extra spacing
            grid.row += 1

            # Header
            grid.add_label("width", column=1)
            grid.add_label("height", column=2)
            grid.row += 1

            # The size as specified in Python
            grid.add_label("python", column=0)
            grid.add_value(py_width_str, column=1)
            grid.add_value(py_height_str, column=2)
            grid.row += 1

            # The component's natural size
            grid.add_label("natural", column=0)
            grid.add_value(
                str(round(self.component_natural_width, 2)), column=1
            )
            grid.add_value(
                str(round(self.component_natural_height, 2)), column=2
            )
            grid.row += 1

            # The component's allocated size
            grid.add_label("allocated", column=0)
            grid.add_value(
                str(round(self.component_allocated_width, 2)), column=1
            )
            grid.add_value(
                str(round(self.component_allocated_height, 2)), column=2
            )
            grid.row += 1

        # Push all of the content to the left. This could be done by aligning
        # the entire Grid, but that would ellipsize some long texts. Instead,
        # add a Spacer into a fifth column, which will take up any unused space.
        grid.add(
            rio.Spacer(height=0),
            column=5,
        )

        # Done
        return grid.as_rio_component()

    def _create_extra_details(
        self,
        debug_details: dict[str, Any],
    ) -> Iterable[rio.Component]:
        grid = DetailsGrid()

        # Custom properties
        #
        # Make sure to skip any which already have custom tailored cells
        for prop_name, prop_value in debug_details.items():
            # Some values have special handling below
            if prop_name in (
                "key",
                "width",
                "height",
                "margin",
                "margin_x",
                "margin_y",
                "margin_left",
                "margin_right",
                "margin_top",
                "margin_bottom",
                "align_x",
                "align_y",
            ):
                continue

            # Display this property
            grid.add_row(prop_name, repr(prop_value))

        if grid.row > 0:
            yield rio.Revealer(
                "More Attributes",
                grid.as_rio_component(),
                # header_style="heading3",
            )

    def _create_layout_revealer(self, target) -> rio.Component:
        return rio.Revealer(
            "Layout",
            layout_preview.LayoutPreview(component=target),
            header_style="heading3",
        )


class DetailsGrid:
    def __init__(self):
        self.grid = rio.Grid(row_spacing=0.5, column_spacing=0.5)
        self.row = 0
        self.max_column = 0

    def add_row(self, label: str, value: str) -> None:
        self.add_label(label, column=0)
        self.add_value(value, column=1, ellipsize=True)
        self.row += 1

    def add_label(
        self,
        text: str,
        *,
        row: int | None = None,
        column: int,
        width: int = 1,
    ) -> None:
        self.add(
            rio.Text(
                text,
                style="dim",
                justify="right",
            ),
            row=row,
            column=column,
            width=width,
        )

    def add_value(
        self,
        value: str,
        *,
        row: int | None = None,
        column: int,
        width: int = 1,
        ellipsize: bool = False,
    ) -> None:
        self.add(
            rio.Text(
                value,
                justify="left",
                width="grow" if ellipsize else "natural",
                wrap="ellipsize" if ellipsize else False,
            ),
            row=row,
            column=column,
            width=width,
        )

    def add(
        self,
        child: rio.Component,
        *,
        row: int | None = None,
        column: int,
        width: int = 1,
        height: int = 1,
    ) -> None:
        if row is None:
            row = self.row

        self.grid.add(child, row, column, width=width, height=height)

        self.max_column = max(self.max_column, column)

    def as_rio_component(self) -> rio.Component:
        return self.grid



from __future__ import annotations

from dataclasses import KW_ONLY
from typing import final

import rio

from .fundamental_component import FundamentalComponent

__all__ = [
    "Switcher",
]


@final
class Switcher(FundamentalComponent):
    """
    Smoothly transitions between components.

    The `Switcher` component is a container which can display one component at a
    time. What makes it useful, is that when you change the `content` attribute,
    rather than instantly swapping the displayed component, it will smoothly
    transition between the two.

    Moreover, whenever the content's size changes, the `Switcher` will
    smoothly resize to match the new size. This means you can use switchers
    to smoothly transition between components of different sizes.

    `content` may also be `None`, in which case the `Switcher` won't display
    anything. This in turn allows you to animate the appearance or disappearance
    of a component, e.g. for a sidebar.


    ## Attributes

    `content`: The component to display inside the switcher. If `None`, the
        switcher will be empty.

    `transition_time`: How many seconds it should take for the switcher to
        transition between components and sizes.


    ## Metadata

    `experimental`: True
    """

    content: rio.Component | None

    _: KW_ONLY

    transition_time: float = 0.35


Switcher._unique_id = "Switcher-builtin"



import functools
import warnings
from typing import *

from .component_meta import ComponentMeta
from .warnings import *

__all__ = ["deprecated", "parameters_renamed", "_remap_kwargs"]


C = TypeVar("C", bound=Union[Callable, ComponentMeta])


def deprecated(*, since: str, description: str):
    def decorator(callable_: C) -> C:
        callable_.__rio_deprecated_since = since  # type: ignore
        callable_.__rio_deprecated_description = description  # type: ignore
        return callable_

    return decorator


def parameters_renamed(old_names_to_new_names: Mapping[str, str]):
    """
    Class/Function decorator that allows it to be called with the old names of
    the given parameters. For example:

        @parameters_renamed({'foo': 'bar'})
        def example_func(bar: int):
            return bar

        example_func(foo=3)  # Ok at runtime, but static type checker complains
    """

    def decorator(callable_: C) -> C:
        if isinstance(callable_, ComponentMeta):
            callable_._deprecated_parameter_names_.update(
                old_names_to_new_names
            )
            return callable_  # type: ignore (wtf?)

        @functools.wraps(callable_)  # type: ignore (wtf?)
        def wrapper(*args, **kwargs):
            _remap_kwargs(callable_.__name__, kwargs, old_names_to_new_names)
            return callable_(*args, **kwargs)  # type: ignore (wtf?)

        return wrapper  # type: ignore (wtf?)

    return decorator


def _remap_kwargs(
    func_name: str,
    kwargs: dict[str, object],
    old_names_to_new_names: Mapping[str, str],
) -> None:
    for old_name, new_name in old_names_to_new_names.items():
        try:
            kwargs[new_name] = kwargs.pop(old_name)
        except KeyError:
            pass
        else:
            warnings.warn(
                f"The {old_name!r} parameter of rio.{func_name} is deprecated; it has been renamed to {new_name!r}",
                RioDeprecationWarning,
            )
